from typing import Any, Dict, List


def sig_show(hub, tree: Dict[str, Any]):
    ...


def sig_process_mod(
    hub,
    ref: str,
    doc: str,
    file: str,
    attributes: List[str],
    functions: Dict[str, Dict[str, Any]],
    variables: Dict[str, Dict[str, Any]],
    classes: Dict[str, Dict[str, Any]],
):
    ...



import json
from typing import Any, Dict, List

__virtualname__ = "json"


def _serialize(d: Dict[str, Any]):
    ret = {}
    for k, v in d.items():
        try:
            json.dumps(v)
            ret[k] = v
        except TypeError:
            ret[k] = str(v)
    return ret


def process_mod(
    hub,
    ref: str,
    doc: str,
    file: str,
    attributes: List[str],
    functions: Dict[str, Dict[str, Any]],
    variables: Dict[str, Dict[str, Any]],
    classes: Dict[str, Dict[str, Any]],
):
    return {
        "ref": ref,
        "doc": doc,
        "file": file,
        "attributes": attributes,
        "functions": {k: _serialize(v) for k, v in functions.items()},
        "variables": {k: _serialize(v) for k, v in variables.items()},
        "classes": {k: _serialize(v) for k, v in classes.items()},
    }


def show(hub, tree: Dict[str, Any]):
    serializable_tree = hub.graph.init.recurse(tree)

    print(json.dumps(serializable_tree))



def __init__(hub):
    hub.pop.sub.add(dyne_name="tree")


def cli(hub):
    hub.pop.config.load(["pop_doc", "pop_tree", "rend"], cli="pop_doc")

    ref = hub.OPT.pop_doc.ref
    split = ref.split(".")
    dyne = split[0]

    if not hasattr(hub, dyne):
        hub.pop.sub.add(dyne_name=dyne)
        should_recurse = len(split) > 3 and not hasattr(hub, ref)
        hub.pop.sub.load_subdirs(getattr(hub, dyne), recurse=should_recurse)

    tree = hub.tree.init.traverse()
    refs = hub.tree.init.refs(tree)

    if ref in refs:
        ret = refs[ref]
    else:
        try:
            ret = tree
            for r in split:
                ret = ret[r]
        except KeyError:
            raise KeyError(f"Reference does not exist on the hub: {ref}")

    print(hub.output[hub.OPT.rend.output].display(ret))



CLI_CONFIG = {
    "output": {
        "source": "rend",
        "default": None,
    },
    "add_sub": {"nargs": "*"},
    "sub": {
        "positional": True,
        "nargs": "?",
    },
    "recurse": {
        "action": "store_true",
    },
    "graph": {},
}

CONFIG = {
    "add_sub": {
        "help": "Add a sub to the hub",
        "default": [],
    },
    "sub": {
        "type": str,
        "help": "The sub on the hub to parse",
        "default": None,
    },
    "recurse": {
        "help": "Load the named sub onto the hub recursively",
        "default": False,
    },
    "graph": {
        "help": "Plugin to use for generating a graph, (I.E. 'simple', 'details', 'json')",
        "default": None,
    },
}

DYNE = {
    "graph": ["graph"],
    "tree": ["tree"],
}



from enum import Enum, IntEnum
from typing import Any, Dict, List

try:
    import networkx as nx
    import matplotlib.cm as cmx
    import matplotlib.colors as colors
    import matplotlib.pyplot as plt

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def __init__(hub):
    # A collection of nodes (vertices) along with identified pairs of nodes (called edges, links, etc).
    hub.graph.networkx.GRAPH = nx.Graph(name="hub")


class NodeType(IntEnum):
    hub = 0
    sub = 1
    plugin = 2
    function = 3
    contract = 6
    variable = 4
    class_ = 5
    unknown = 7


def add_node(hub, ref: str, node_type: NodeType = NodeType.unknown, **attrs):
    builder = []
    for name in ref.split("."):
        parent = ".".join(builder)
        cur_ref = ".".join(builder + [name])
        if cur_ref not in hub.graph.networkx.GRAPH.nodes:
            hub.graph.networkx.GRAPH.add_node(
                cur_ref, name=name, node_type=NodeType.sub
            )
        hub.graph.networkx.GRAPH.add_edge(parent, cur_ref)
        builder.append(name)

    # The last thing in the list gets all the attrs
    if attrs:
        hub.graph.networkx.GRAPH.add_node(
            cur_ref, name=name, node_type=node_type, **attrs
        )


def add_edge(
    hub, ref1: str, ref2: str, node_type: NodeType = NodeType.unknown, **ref2_attrs
):
    hub.graph.networkx.add_node(ref2, node_type=node_type, **ref2_attrs)
    hub.graph.networkx.GRAPH.add_edge(ref1, ref2)


def process_mod(
    hub,
    ref: str,
    doc: str,
    file: str,
    attributes: List[str],
    functions: Dict[str, Dict[str, Any]],
    variables: Dict[str, Dict[str, Any]],
    classes: Dict[str, Dict[str, Any]],
):
    hub.graph.networkx.add_node(
        ref, doc=doc, file=file, attributes=attributes, node_type=NodeType.plugin
    )

    for name, function in functions.items():
        if any(function["contracts"].get(c) for c in ("pre", "call", "post")):
            previous = None
            for c in (
                function["contracts"]["pre"]
                + function["contracts"]["call"]
                + [function["ref"]]
                + function["contracts"]["post"]
            ):
                if c == function["ref"]:
                    node_type = NodeType.function
                else:
                    node_type = NodeType.contract
                hub.graph.networkx.GRAPH.add_node(
                    c, name=c.split(".")[-1], node_type=node_type
                )
                if previous is None:
                    hub.graph.networkx.GRAPH.add_edge(ref, c, doc=doc)
                else:
                    hub.graph.networkx.GRAPH.add_edge(previous, c)
                previous = c
        else:
            hub.graph.networkx.add_edge(
                ref, function["ref"], doc=doc, node_type=NodeType.function
            )

    for name, variable in variables.items():
        hub.graph.networkx.add_edge(
            ref,
            variable["ref"],
            type=variable["type"],
            value=variable["value"],
            node_type=NodeType.variable,
        )

    for name, cls in classes.items():
        hub.graph.networkx.add_edge(ref, cls["ref"], node_type=NodeType.class_)


def show(hub, tree: Dict[str, Any]):
    hub.graph.init.recurse(tree)
    hub.graph.networkx.GRAPH.add_node("", name="hub", node_type=NodeType.hub)

    # All nodes should point to themselves down the line
    graph: nx.GRAPH = hub.graph.networkx.GRAPH
    # from pprint import pprint
    # print(dict(graph.adjacency()))

    fig = plt.figure()
    ax = fig.add_subplot(111)
    pos_nodes = nx.kamada_kawai_layout(graph)
    cmap = plt.cm.get_cmap("Set2")
    color_norm = colors.Normalize(vmin=min(NodeType), vmax=max(NodeType))
    scalar_map = cmx.ScalarMappable(norm=color_norm, cmap=cmap)
    for item in NodeType:
        ax.plot(
            [0], [0], color=scalar_map.to_rgba(item.value), label=item.name.strip("_")
        )
    nx.draw(
        graph,
        pos_nodes,
        vmin=min(NodeType),
        vmax=max(NodeType),
        node_color=[
            v.value for v in nx.get_node_attributes(graph, "node_type").values()
        ],
        cmap=cmap,
    )
    plt.legend()
    names = nx.get_node_attributes(graph, "name")
    nx.draw_networkx_labels(graph, pos_nodes, labels=names)
    plt.legend()
    fig.tight_layout()
    plt.show()



from typing import Any, Dict


def __init__(hub):
    hub.graph.GRAPH = None


def is_mod(hub, d: Dict[str, Any]) -> bool:
    """
    Return True if the dictionary looks like the tree's representation of a plugin, else false
    """
    if not all(
        att in d
        for att in (
            "ref",
            "doc",
            "file",
            "attributes",
            "classes",
            "functions",
            "variables",
        )
    ):
        return False
    if not isinstance(d["functions"], dict):
        return False
    if not isinstance(d["variables"], dict):
        return False
    return True


def recurse(hub, d: Dict[str, Any]):
    ret = {}
    for k, v in d.items():
        if isinstance(v, dict):
            if hub.graph.init.is_mod(v):
                ret[k] = getattr(hub.graph, hub.graph.GRAPH).process_mod(**v)
            else:
                ret[k] = hub.graph.init.recurse(v)
        else:
            ret[k] = v
    return ret


def show(hub, tree: Dict[str, Any]):
    getattr(hub.graph, hub.graph.GRAPH).show(tree)



CLI_CONFIG = {
    "output": {
        "source": "rend",
        "default": None,
    },
    "ref": {"positional": True},
}

CONFIG = {
    "ref": {
        "type": str,
        "help": "The sub on the hub to parse",
        "default": None,
    },
}

DYNE = {"doc": ["doc"]}



# These are just for fun when viewing the graph


def pre(hub, ctx):
    ...


def pre_show(hub, ctx):
    ...


def call(hub, ctx):
    return ctx.func(*ctx.args, **ctx.kwargs)


def call_show(hub, ctx):
    return ctx.func(*ctx.args, **ctx.kwargs)


def post(hub, ctx):
    return ctx.ret


def post_show(hub, ctx):
    return ctx.ret



from pop.hub import Hub, Sub
from dict_tools import data as data_
from typing import Any, Dict

__func_alias__ = {"format_": "format"}


def __init__(hub: Hub):
    hub.pop.sub.add(dyne_name="rend")
    hub.pop.sub.add(dyne_name="graph")
    hub.pop.sub.load_subdirs(hub.tree)


def cli(hub: Hub):
    hub.pop.config.load(["pop_tree", "rend"], cli="pop_tree")

    print_opt = str(hub.OPT.pop_tree.sub).upper() == "OPT"
    if hub.OPT.pop_tree.sub:
        if print_opt:
            hub.pop.config.load(
                ["tree", "rend"] + list(hub.OPT.pop_tree.add_sub), cli="tree"
            )
        else:
            hub.pop.sub.add(dyne_name=hub.OPT.pop_tree.sub)
            if hub.OPT.pop_tree.recurse:
                hub.pop.sub.load_subdirs(
                    getattr(hub, hub.OPT.pop_tree.sub), recurse=True
                )

    for dyne in hub.OPT.pop_tree.add_sub:
        hub.pop.sub.add(dyne_name=dyne)

    if print_opt:
        result = hub.OPT
    else:
        result = hub.tree.init.traverse()
        if hub.OPT.pop_tree.sub:
            result = {hub.OPT.pop_tree.sub: result[hub.OPT.pop_tree.sub]}

    if hub.OPT.pop_tree.graph:
        hub.graph.GRAPH = hub.OPT.pop_tree.graph
    else:
        # Find the first plugin that was loaded for graphing
        loaded_mods = hub.graph._loaded
        if "simple" in loaded_mods:
            hub.graph.GRAPH = "simple"
        else:
            iter_mods = iter(hub.graph._loaded)
            hub.graph.GRAPH = next(iter_mods)
            if hub.graph.GRAPH == "init":
                hub.graph.GRAPH = next(iter_mods)

    hub.graph.init.show(result)


def recurse(hub: Hub, sub: Sub, ref: str = None) -> Dict[str, Any]:
    """
    Find all of the loaded subs in a Sub. I.E:
        pprint(hub.pop.tree.recurse(hub.pop))
    :param hub: The redistributed pop central hub
    :param sub: The pop object that contains the loaded module data
    :param ref: The current reference on the hub
    """
    sub_name = sub._dyne_name
    if sub_name:
        if ref:
            ref = f"{ref}.{sub_name}"
        else:
            ref = sub_name
    ret = data_.NamespaceDict()
    for loaded in sorted(sub._subs):
        loaded_ref = f"{ref}.{loaded}"
        loaded_sub: Sub = getattr(sub, loaded)
        if not (
            getattr(loaded_sub, "_virtual", False)
            and getattr(loaded_sub, "_sub_virtual", True)
        ):
            # Bail early if the sub's virtual isn't True
            continue
        recursed_sub = hub.tree.init.recurse(loaded_sub, ref=loaded_ref)

        for mod in sorted(loaded_sub._loaded):
            loaded_mod = getattr(loaded_sub, mod)
            recursed_sub[mod] = hub.tree.mod.parse(
                loaded_mod, ref=f"{ref}.{loaded}.{mod}"
            )

        if recursed_sub:
            ret[loaded] = recursed_sub

    return ret


def traverse(hub: Hub) -> Dict[str, Any]:
    """
    :param hub: The redistributed pop central hub
    :return: A dictionary representation of all the subs on the hub. I.E:
        pprint(hub.pop.tree.traverse())
    """
    root = data_.NamespaceDict()
    for sub in sorted(hub._subs):
        loaded_sub = getattr(hub, sub)
        root[sub] = hub.tree.init.recurse(loaded_sub)

        for mod in sorted(loaded_sub._loaded):
            loaded_mod = getattr(loaded_sub, mod)
            root[sub][mod] = hub.tree.mod.parse(loaded_mod, ref=f"{sub}.{mod}")
    return root


def refs(hub, tree: Dict[str, Any]) -> Dict[str, Any]:
    """
    Return all the references available on the hub by reference first
    """
    ret = {}

    def _get_refs(t: Dict[str, Any]):
        for k, v in t.items():
            if isinstance(v, Dict):
                _get_refs(v)
            elif k == "ref":
                ret[t["ref"]] = t

    _get_refs(tree)

    return ret



import inspect
import textwrap
from pop.contract import Contracted as Contract
from pop.hub import Hub
from pop.loader import LoadedMod
from collections import OrderedDict
from typing import Any, Dict, List

UNKNOWN_REF = "< unknown ref >"


def get_ref(hub: Hub, mod: LoadedMod) -> str:
    """
    Try to find a reference on the hub for the given mod
    """
    try:
        sister_func: Contract = next(iter(mod._funcs.values()))
        return sister_func.ref
    except StopIteration:
        return "asdf"


def serialize_signature(hub: Hub, signature: inspect.Signature):
    ret = OrderedDict()
    for p in signature.parameters:
        param: inspect.Parameter = signature.parameters[p]
        ret[param.name] = {}
        if param.default is not inspect._empty:
            ret[param.name]["default"] = param.default
        if param.annotation is not inspect._empty:
            ret[param.name]["annotation"] = str(param.annotation)

    final_ret = {"parameters": ret}
    if signature.return_annotation is not inspect._empty:
        final_ret["return_annotation"] = signature.return_annotation
    return final_ret


def funcs(hub: Hub, mod: LoadedMod, ref: str) -> List[str] or Dict[str, str]:
    """
    Find all of the loaded functions in a pop plugin. I.E:
        pprint(hub.pop.tree.funcs(hub.pop.tree))
    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    :return: A Dictionary of loaded modules names mapped to a list of their functions
    """
    funcs = sorted(mod._funcs.keys())
    ret = {}
    for f in funcs:
        contract: Contract = mod._funcs[f]
        func_info = {
            "ref": f"{ref}.{f}",
            "doc": textwrap.dedent(str(contract.func.__doc__ or "")).strip("\n"),
            "contracts": {
                contract_type: [f"{c.ref}.{c.func.__name__}" for c in contracts]
                for contract_type, contracts in contract.contract_functions.items()
            },
        }
        func_info.update(hub.tree.mod.serialize_signature(contract.signature),)
        ret[f] = func_info
    return ret


def data(hub: Hub, mod: LoadedMod, ref: str) -> List[str] or Dict[str, str]:
    """
    Find all of the loaded data in a pop plugin. I.E:
        pprint(hub.pop.tree.data(hub.pop.tree))
    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    """
    datas = sorted(x for x in mod._vars if x.isupper() and not x.startswith("_"))
    ret = {}
    for d_name in datas:
        d = mod._vars[d_name]
        data_info = {
            "ref": f"{ref}.{d_name}",
            "type": d.__class__.__name__,
            "value": d,
        }
        ret[d_name] = data_info
    return ret


def types(hub: Hub, mod: LoadedMod, ref: str) -> List[str] or Dict[str, str]:
    """
    Find all of the loaded types in a pop plugin. I.E:
        pprint(hub.pop.tree.types(hub.pop.tree))
    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    """
    classes = sorted(x for x in mod._classes if not x.startswith("_"))
    ret = {}
    for class_name in classes:
        c = mod._classes[class_name]
        signature = inspect.signature(c.__init__)
        class_info = {
            "ref": f"{ref}.{class_name}",
            "doc": textwrap.dedent((c.__doc__ or "")).strip("\n"),
            "signature": f"{hub.tree.mod.serialize_signature(signature)}",
        }
        ret[class_name] = class_info
    return ret


def parse(hub: Hub, mod: LoadedMod, ref: str) -> Dict[str, Any]:
    """
    Parse a loaded mod object

    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    """
    mod_functions = hub.tree.mod.funcs(mod, ref)
    mod_variables = hub.tree.mod.data(mod, ref)
    ret = {
        "ref": ref,
        "doc": (mod._attrs.get("__doc__") or "").strip(),
        "file": getattr(mod, "__file__", None),
        "attributes": sorted(
            [a for a in mod._attrs if not (a.startswith("__") and a.endswith("__"))]
        ),
        "classes": hub.tree.mod.types(mod, ref),
        "functions": mod_functions,
        "variables": mod_variables,
    }
    return ret



from collections import namedtuple
from typing import Any, Dict, List

__virtualname__ = "simple"

Plugin = namedtuple("Plugin", ("functions", "variables"))


def __virtual__(hub):
    return (
        hasattr(hub, "output"),
        "'rend' needs to be installed in the python environment to use this plugin",
    )


def _keys(d: Dict[str, Any]) -> List[str]:
    return [k for k in d.keys()]


def process_mod(
    hub,
    ref: str,
    doc: str,
    file: str,
    attributes: List[str],
    functions: Dict[str, Dict[str, Any]],
    variables: Dict[str, Dict[str, Any]],
    classes: Dict[str, Dict[str, Any]],
):
    return Plugin(functions=_keys(functions), variables=_keys(variables))


def show(hub, tree: Dict[str, Any]):
    simple_tree = hub.graph.init.recurse(tree)

    outputter = getattr(hub, f"output.{hub.OPT.rend.output}.display")
    rendered = outputter(simple_tree)
    print(rendered)
